/********************************************
 * Author: Andrey_B (Home Automation http://www.ab-log.ru), Guido Socher (TCP/IP stack)
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * Home Automation universal Ethernet module
 * Remote Control and Configuration
 * Port: IN/OUT/PWM/ADC support
 * HTTP interface 
 *
 * Chip type: Atmega328 with ENC28J60
 *
 * !    ,
 *   ,    "aio"
 *    .
 *********************************************/
#include <avr/io.h>
#include <string.h>
#include <stdlib.h>
#include "websrv_help_functions.h"
#include "ip_arp_udp_tcp.h"
#include "enc28j60.h"
#include "timeout.h"
#include "tcp_state.h"
#include "net.h"

// For sprintf_P
#include <stdio.h>
// For EEPROM
#include <avr/eeprom.h> 
// For Debouncing
#include <avr/interrupt.h>
// For DHT11/DHT22 humidity sensor
#include "hw_dht.h"

// Firmware version
#define MEGADFWV "3.21"

// Default MAC address
//static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
uint8_t mymac[6] = {0x54,0x55,0xc0,0xa8,0x00,0x0e};
// Default IP address
uint8_t myip[4] = {192,168,0,14};

// Default Netmask
static uint8_t netmask[4] = {255,255,255,0};
// Default HTTP port
#define HTTPPORT 80
// MAC resolving defenition
#define TRANS_NUM_GWMAC 1
static uint8_t gwmac[6];
#define TRANS_NUM_WEBMAC 2
//static uint8_t www_gwmac[6];
static uint8_t www_gwmac2[6];

//uint8_t tcp_client_state;

// EEPROM 
// EEMEM IP address
#define ee_ip_addr	0x03F6
// FW Update flag
#define ee_fw_up	0x03F5

//uint8_t EEMEM ee_ip_addr[4];
// IP address
uint8_t _ip_addr[4];
// EEMEM server IP address
uint8_t EEMEM ee_sip_addr[4];
// EEMEM Gateway IP address
uint8_t EEMEM ee_gw_addr[4];
// server IP address
uint8_t _sip_addr[4];
//uint8_t _www_ip[4]; // dns will fill this
// EEMEM server script (MAX 14 bytes!)
uint8_t EEMEM ee_spt[15];
// EEMEM password (MAX 5 bytes!)
uint8_t EEMEM ee_pwd[6];
// server script
char _spt[16];
// password
static char password[7];
// config preset
uint8_t EEMEM ee_preset;
static uint8_t _preset;
// alarm temp
uint8_t EEMEM ee_alarm_temp;
static uint8_t _alarm_temp;

// Temp check flag
uint8_t EEMEM ee_temp_check;
uint8_t temp_check_flag = 0;

// EEMEM ab-srv ID (MAX 5 bytes!)
uint8_t EEMEM ee_mdid[5];
char _mdid[6];

// Available IO
#define IO_SIZE 16
// MegaD-328-7I7O Ver 4-5 sequence
char *aio[IO_SIZE] = {"C5", "C4", "C3", "C2", "C1", "C0", "D7", "B1", "D0", "D1", "D3", "D4", "D5", "D6", "A6", "A7"};
// MegaD-328-7I7O Ver 1-3 sequence
//char *aio[IO_SIZE] = {"C5", "D0", "C4", "D1", "C3", "C2", "D3", "B1", "D7", "D6", "D5", "C0", "D4", "C1", "A6", "A7"};

// EEMEM Input port cmd
uint8_t EEMEM ee_cmd[IO_SIZE][23];
// EEMEM Input port cmd
uint8_t EEMEM ee_eth_cmd[IO_SIZE][16];

// Initializing buffer for HTTP
#define BUFFER_SIZE 800
static uint8_t buf[BUFFER_SIZE+1];
// Buffer for URL parameters
static char gStrbuf[25];

// EEPROM
//Port type
uint8_t EEMEM ee_port_type[IO_SIZE];
static uint8_t _port_type[IO_SIZE];

//Port default state
uint8_t EEMEM ee_port_d[IO_SIZE];
static uint8_t _port_d[IO_SIZE];

//Port default mode (Switch/PWM)
uint8_t EEMEM ee_port_m[IO_SIZE];
static uint8_t _port_m[IO_SIZE];

//Port misc param 
//       
//uint16_t EEMEM ee_port_misc[IO_SIZE];
uint16_t EEMEM ee_port_misc[IO_SIZE];
static uint16_t _port_misc[IO_SIZE];
static uint16_t port_misc_cur[IO_SIZE];

//    
uint8_t EEMEM ee_srv_loop;
uint32_t srv_loop_timer = 0;

#define SRV_LOOP_TIMEOUT 12000 // 12000 * 5ms = 60 sec


// Port act status array
static uint8_t port_act[IO_SIZE];

//Port input counter
uint16_t port_cnt[IO_SIZE];

// Service
char temp[42];
char shost[20];
uint8_t i;
// Reset flag
uint8_t reset_flag = 0;
// ADC check flag
uint8_t adc_check_flag = 0;
uint8_t cur_input;
//static uint8_t cur_input2[IO_SIZE];
//       
static uint16_t srv_timeout[IO_SIZE];
static uint16_t srv_timeout_act[IO_SIZE];

static uint8_t cmd_delay_p[IO_SIZE];

// ARP 
static uint16_t arp_timeout = 0;
// If is_set gateway IP
static uint8_t gw_set = 0;
//    TCP-.     
//       
//static int8_t send_eth_flag = -1;
//   ,   TCP-   
static int8_t send_eth_flag2 = -1;

// Debouncing
char input_state;      //Debounced and inverted key state:
char input_press;      //Key Press Detected  
char input_release;      //Key Release Detected  
char input_state2;      //Debounced and inverted key state:
char input_press2;      //Key Press Detected  
char input_release2;      //Key Release Detected  
char input_state3;      //Debounced and inverted key state:
char input_press3;      //Key Press Detected  
char input_release3;      //Key Release Detected  

// DHT vars
int8_t dht11_temperature = 0;
int8_t dht11_humidity = 0;
uint8_t volatile isr_flag=0;

int8_t port_num = 0;
char port_letter;
int8_t dht_type;

// function for software reset of AVR
#define RESET() (((void(*)(void))(char *)0x0000)()) 

// set output to VCC, red LED off
#define LEDON PORTB|=(1<<PORTB0)
// set output to GND, red LED on
#define LEDOFF PORTB&=~(1<<PORTB0)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB0)

uint32_t adc_timer = 0;
uint32_t temp_timer = 0;

typedef void (*AppPtr_t) (void);
AppPtr_t Bootloader = (AppPtr_t)0x03800;

//#define LEDON2 PORTD|=(1<<PORTD7)

uint16_t readchiptemp(void)
{
	ADMUX = (3 << REFS0) | (8 << MUX0); // 1.1V REF, channel#8 is temperature
	ADCSRA |= (1 << ADEN) | (6 << ADPS0);       // enable the ADC div64
	_delay_ms(20);
	ADCSRA |= (1 << ADSC);      // Start the ADC

	while (ADCSRA & (1 << ADSC));       // Detect end-of-conversion
	// The offset of 324.31 could be wrong. It is just an indication.
	// The returned temperature is in degrees Celcius.
	//return (ADCW - 324.31) / 1.22;
	return (ADCW * 100 - 32431) / 122;
} 

uint16_t http200ok(void)
{
	return(fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\nPragma: no-cache\r\n\r\n")));
}

uint8_t verify_password(char *str)
{
	// the first characters of the received string are
	// a simple password/cookie:
	if (strncmp(password,str,strlen(password))==0)
	return(1);
	return(0);
}

int8_t analyse_get_url(char *str)
{
	uint8_t loop=15;
	// the first slash:
	if ( *str == '/' )
	{ str++; }
	else
	{ return(-1); }

	if ( strncmp("favicon.ico", str, 11) == 0 )
	{ return(2); }
	// the password:
	if (verify_password(str)==0)
	{ return(-1); }

        // move forward to the first space or '/'
	while(loop)
	{
		if(*str==' ')
		{
			// end of url and no slash after password:
			return(-2);
                }

		if(*str=='/')
		{
			// end of password
			loop=0;
			continue;
		}

		str++;
		loop--; // do not loop too long
	}

        return(-3);
}

static uint8_t decode_ip(char *in,uint8_t *out)
{
	//uint8_t i;
	//char tmp[20];
	strncpy(temp,in,sizeof(temp));
	char *dig;
	dig=strtok(temp,".");

	for(i=0 ; i<4 && dig ;i++,dig=strtok(NULL,"."))
	out[i]=(uint8_t)strtoul(dig,NULL,10);

	return i;
}

// ,      
void port_execute(char *srv_cmd, uint8_t c_input, uint8_t cmd_shift)
{
	char p_num[3] = "FF";
	//char pwm[3] = "FFF";
	char pwm[3];
	uint8_t cnt=0;
	uint8_t p_flag=0;
	char port_letter;
	int8_t port_num;

	//uint16_t cmd_delay;
	char cmd_delay[5];

	for ( i = 0; i < strlen(srv_cmd); i++ )
	{
		if ( srv_cmd[i] == '\n' )
		{
			p_flag = 0;
			p_num[0] = 'F';
			p_num[1] = 'F';
			pwm[0] = *"";
			pwm[1] = *"";
			pwm[2] = *"";
			cnt = 0;
		}
		else if ( srv_cmd[i] == ':' && p_flag != 3 )
		p_flag = 1;
		else if ( srv_cmd[i] == ';' )
		p_flag = 0;
		else if ( srv_cmd[i] == 'p' )
		p_flag = 2;
		else if ( srv_cmd[i] == 'a' )
		p_flag = 3;
		else
		{
			if ( p_flag == 0 )
			{
				if ( p_num[0] != 'F' )
				p_num[1] = srv_cmd[i];
				else
				p_num[0] = srv_cmd[i];
			}
			else if ( p_flag == 1 )
			{
				if ( p_num[1] == 'F' )
				p_num[1] = '\0';
				port_letter = aio[atoi(p_num)][0];
				port_num = atoi(&aio[atoi(p_num)][1]);

				if (_port_type[atoi(p_num)] == 1 )
				{
					if ( _port_m[atoi(p_num)] == 1 )
					{
						uint8_t j = i + 1;
						if ( j >= strlen(srv_cmd) || srv_cmd[j] == ':' || srv_cmd[j] == ';' )
						{
							pwm[cnt] = srv_cmd[i];
							if ( port_letter == 'D' )
							{
								if ( port_num == 6 )
								OCR0A = atoi(pwm);
								else if ( port_num == 5 )
								OCR0B = atoi(pwm);
								else if ( port_num == 3 )
								OCR2B = atoi(pwm);
							}

							p_num[0] = 'F';
							p_num[1] = 'F';
							//pwm[0] = 'F';
							//pwm[1] = 'F';
							//pwm[2] = 'F';
							pwm[0] = *"";
							pwm[1] = *"";
							pwm[2] = *"";
							cnt = 0;
						}
						else
						{
							pwm[cnt] = srv_cmd[i];
							cnt++;
						}
						
					}
					else
					{
						// ON
						if ( srv_cmd[i] == '1' )
						{
							if ( port_letter == 'D' )
							PORTD|= (1<<port_num);
							else if ( port_letter == 'C' )
							PORTC|= (1<<port_num);
							else if ( port_letter == 'B' )
							PORTB|= (1<<port_num);
						}
						// OFF
						else if ( srv_cmd[i] == '0' )
						{
							if ( port_letter == 'D' )
							PORTD &= ~(1<<port_num);
							else if ( port_letter == 'C' )
							PORTC &= ~(1<<port_num);
							else if ( port_letter == 'B' )
							PORTB &= ~(1<<port_num);
						}
						// Toggle
						else if ( srv_cmd[i] == '2' )
						{
							if ( port_letter == 'D' )
							PORTD ^= (1<<port_num);
							else if ( port_letter == 'C' )
							PORTC ^= (1<<port_num);
							else if ( port_letter == 'B' )
							PORTB ^= (1<<port_num);
						}

						p_num[0] = 'F';
						p_num[1] = 'F';
					}
				}
			}
			else if ( p_flag == 2 )
			{
				uint8_t j = i + 1;
				cmd_delay[cnt] = srv_cmd[i];
				cnt++;
				if ( j >= strlen(srv_cmd) || srv_cmd[j] == ':' || srv_cmd[j] == ';' )
				{
					srv_timeout_act[c_input] = atoi(cmd_delay) * 20;
					cmd_delay_p[c_input] = j + 1 + cmd_shift;
					memset(cmd_delay, 0, 5);
					break;

				}
				
			}
			else if ( p_flag == 3 )
			{
				uint8_t j;

				for ( j = 0; j < IO_SIZE; j++ )
				{
					port_letter = aio[j][0];
					port_num = atoi(&aio[j][1]);

					if ( _port_type[j] == 1 )
					{
						// ON
						if ( srv_cmd[i] == '1' )
						{
							if ( port_letter == 'D' )
							PORTD|= (1<<port_num);
							else if ( port_letter == 'C' )
							PORTC|= (1<<port_num);
							else if ( port_letter == 'B' )
							PORTB|= (1<<port_num);
						}
						// OFF
						else if ( srv_cmd[i] == '0' )
						{
							if ( port_letter == 'D' )
							PORTD &= ~(1<<port_num);
							else if ( port_letter == 'C' )
							PORTC &= ~(1<<port_num);
							else if ( port_letter == 'B' )
							PORTB &= ~(1<<port_num);
						}
						// Toggle
						else if ( srv_cmd[i] == '2' )
						{
							if ( port_letter == 'D' )
							PORTD ^= (1<<port_num);
							else if ( port_letter == 'C' )
							PORTC ^= (1<<port_num);
							else if ( port_letter == 'B' )
							PORTB ^= (1<<port_num);
						}

					}
				}
			}


		}
	}

}

// --- void browserresult_callback(uint8_t statuscode,uint16_t datapos, uint16_t len)
void browserresult_callback(uint16_t statuscode,uint16_t datapos __attribute__((unused)), uint16_t len __attribute__((unused)), uint8_t c_input)
{
	srv_timeout[c_input] = 0;

	uint16_t i=0;
	uint8_t j=0;
	uint8_t k=0;

	char * srv_rep;
	srv_rep = (char *)&(buf[datapos]);
	char srv_cmd[30];

	for ( i = 0; i < 30; i++ )
	srv_cmd[i] = *"";

	// ,           200 
	if ( len > 220 )
	len = 220;


	if (statuscode==200)
	{
		for ( i = 0; i < len; i++ )
		{
			//   HTTP-
			if ( (*srv_rep == '\r' || *srv_rep == '\n') && k != 2 )
			{
				srv_rep++;
				if ( *srv_rep == '\n' )
				k++;
				continue;
			}

			if ( k == 2 )
			{
				srv_cmd[j] = *srv_rep;
				j++;
			}

			if ( k != 2 )
			k = 0;

			srv_rep++;
		}

		if (strlen(srv_cmd) > 0)
		port_execute(srv_cmd, 100, 0);
	}
	//     200,     
	else
	{
		memset(temp, 0, sizeof(temp));
		eeprom_read_block (temp, &ee_cmd[c_input], 23);
		if ( temp[0] != (char)255 && strlen(temp) > 0 ) 
		port_execute(temp, c_input, 0);

	}

	send_eth_flag2 = 1;
}


void browserresult_callback_empty(uint16_t statuscode,uint16_t datapos __attribute__((unused)), uint16_t len __attribute__((unused)), uint8_t c_input __attribute__((unused)))
{
	if (statuscode==200)
	{}
	send_eth_flag2 = -1;
}


void arpresolver_result_callback(uint8_t *ip __attribute__((unused)),uint8_t reference_number,uint8_t *mac){
        uint8_t i=0;
        if (reference_number==TRANS_NUM_GWMAC){
                // copy mac address over:
                while(i<6){gwmac[i]=mac[i];i++;}
        }
	/*
        if (reference_number==TRANS_NUM_WEBMAC){
                // copy mac address over:
                while(i<6){www_gwmac[i]=mac[i];i++;}
        }
	*/
}

void arpresolver_result_callback2(uint8_t *ip __attribute__((unused)),uint8_t reference_number,uint8_t *mac){
        uint8_t i=0;
	      // copy mac address over:
        if (reference_number==TRANS_NUM_WEBMAC){
                while(i<6){www_gwmac2[i]=mac[i];i++;}
	}
}


ISR(TIMER1_COMPA_vect)            //every 5ms
{
	char i;
	static char ct0 = 0xFF, ct1 = 0xFF;  // 8 * 2bit counters
	static char ct2 = 0xFF, ct3 = 0xFF;  // 8 * 2bit counters
	static char ct4 = 0xFF, ct5 = 0xFF;  // 8 * 2bit counters

	uint8_t j;

	if ( adc_check_flag == 1 )
	adc_timer++;

	if ( temp_check_flag > 0 )
	temp_timer++;

	if ( srv_loop_timer > 2 )
	srv_loop_timer--;

	//      
	for ( j = 0; j < IO_SIZE; j++ )
	{
		//   ,   
		if ( srv_timeout[j] > 0 )
		{
			srv_timeout[j]--;
			//     ,   -    
			if ( srv_timeout[j] == 0 )
			srv_timeout_act[j] = 1;
		}

		if ( srv_timeout_act[j] > 2 )
		srv_timeout_act[j]--;
	}

	i = ~PIND;              // read keys (low active)
	i ^= input_state;            // key changed ?
	ct0 = ~(ct0 & i);            // reset or count ct0
	ct1 = ct0 ^ (ct1 & i);       // reset or count ct1
	i &= ct0 & ct1;              // count until roll over ?
	input_release |= input_state & i;  // 1->0: key release detect 
	input_state ^= i;            // then toggle debounced state
	input_press |= input_state & i;   // 0->1: key press detect

	i = ~PINC;              // read keys (low active)
	i ^= input_state2;            // key changed ?
	ct2 = ~(ct2 & i);            // reset or count ct0
	ct3 = ct2 ^ (ct3 & i);       // reset or count ct1
	i &= ct2 & ct3;              // count until roll over ?
	input_release2 |= input_state2 & i;  // 1->0: key release detect 
	input_state2 ^= i;            // then toggle debounced state
	input_press2 |= input_state2 & i;   // 0->1: key press detect

	i = ~PINB;              // read keys (low active)
	i ^= input_state3;            // key changed ?
	ct4 = ~(ct4 & i);            // reset or count ct0
	ct5 = ct4 ^ (ct5 & i);       // reset or count ct1
	i &= ct4 & ct5;              // count until roll over ?
	input_release3 |= input_state3 & i;  // 1->0: key release detect 
	input_state3 ^= i;            // then toggle debounced state
	input_press3 |= input_state3 & i;   // 0->1: key press detect

	isr_flag = 1;
} 

char get_key_press( char input_mask )
{
	cli();
	input_mask &= input_press;                // read key(s)
	input_press ^= input_mask;                // clear key(s)
	sei();
	return input_mask;
}

char get_key_release( char input_mask )
{
	cli();
	input_mask &= input_release;                // read key(s)
	input_release ^= input_mask;                // clear key(s)
	sei();
	return input_mask;
}

char get_key_press2( char input_mask )
{
	cli();
	input_mask &= input_press2;                // read key(s)
	input_press2 ^= input_mask;                // clear key(s)
	sei();
	return input_mask;
}

char get_key_release2( char input_mask )
{
	cli();
	input_mask &= input_release2;                // read key(s)
	input_release2 ^= input_mask;                // clear key(s)
	sei();
	return input_mask;
}

char get_key_press3( char input_mask )
{
	cli();
	input_mask &= input_press3;                // read key(s)
	input_press3 ^= input_mask;                // clear key(s)
	sei();
	return input_mask;
}

char get_key_release3( char input_mask )
{
	cli();
	input_mask &= input_release3;                // read key(s)
	input_release3 ^= input_mask;                // clear key(s)
	sei();
	return input_mask;
}

int main(void)
{
        uint16_t plen;
	uint16_t dat_p;
	int8_t cmd;

	//uint8_t k;
	uint8_t _eth_addr[4];

	int8_t avr_port;
	int8_t avr_port_port;
	int8_t avr_port_pin;
	//int8_t port_num = 0;

	//char temp[25];
	//char port_letter;

	uint16_t plen_start = 0;

	CLKPR=(1<<CLKPCE); // change enable
	CLKPR=0; // "no pre-scaler"
	_delay_loop_1(0); // 60us

	// Reading EEPROM IP address
	//eeprom_read_block ((void *)_ip_addr, (const void *)&ee_ip_addr,4);
	eeprom_read_block ((void *)_ip_addr, (const void *)ee_ip_addr,4);
	if ( _ip_addr[0] != 255 )
	{
		for ( i = 0; i < 4; i++ )
		{
			myip[i] = _ip_addr[i];
			mymac[i+2] = _ip_addr[i];
		}
	}

	/*initialize enc28j60*/
	enc28j60Init(mymac);
	enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
	_delay_loop_1(0); // 60us

	/* Magjack leds configuration, see enc28j60 datasheet, page 11 */
	// LEDB=yellow LEDA=green
	// 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
	enc28j60PhyWrite(PHLCON,0x476);

	eeprom_read_block (_port_type, &ee_port_type, IO_SIZE);
	eeprom_read_block (_port_d, &ee_port_d, IO_SIZE);
	eeprom_read_block (_port_m, &ee_port_m, IO_SIZE);
	eeprom_read_block (_port_misc, &ee_port_misc, IO_SIZE * 2);

	_preset = eeprom_read_byte(&ee_preset);
	_alarm_temp = eeprom_read_byte(&ee_alarm_temp);

	eeprom_read_block (password, &ee_pwd, 6);
	// Default password
	if ( password[0] == (char)255 )
	strcpy(password, "sec");

	//strcpy(password, "sec");

	DDRB|= (1<<DDB0); // enable PB0, LED as output 

	// Enable ADC
	// Set ADC prescalar
	ADCSRA |= (1 << ADPS2) | (1 << ADPS1);
	// Set ADC reference to AVCC 
	ADMUX |= (1 << REFS0);
	// Left adjust ADC result to allow easy 8 bit reading 
	//ADMUX |= (1 << ADLAR); 
	// Enable ADC 
	ADCSRA |= (1<<ADEN);
	// Start A2D Conversions 
	//ADCSRA |= (1<<ADSC);

	for ( i = 0; i < IO_SIZE; i++ )
	{
		port_letter = aio[i][0];
		port_num = atoi(&aio[i][1]);
		srv_timeout[i] = 0;
		srv_timeout_act[i] = 0;
		port_misc_cur[i] = 0;

		if (_port_type[i] == 1 )
		{
			if ( port_letter == 'D' )
			{
				DDRD |= (1<<port_num);
				if (_port_d[i] == 1 )
				PORTD|= (1<<port_num);
				else
				PORTD &= ~(1<<port_num);

				//  
				if ( port_num == 6 )
				TCCR0A &= ~(1 << COM0A1);
				else if ( port_num == 5 )
				TCCR0A &= ~(1 << COM0B1);
				else if ( port_num == 3 )
				TCCR2A &= ~(1 << COM2B1);

			}
			else if ( port_letter == 'C' )
			{
				DDRC |= (1<<port_num);
				if (_port_d[i] == 1 )
				PORTC|= (1<<port_num);
				else
				PORTC &= ~(1<<port_num);
			}
			else if ( port_letter == 'B' )
			{
				DDRB |= (1<<port_num);
				if (_port_d[i] == 1 )
				PORTB|= (1<<port_num);
				else
				PORTB &= ~(1<<port_num);
			}

			//  
			if ( _port_m[i] == 1 )
			{
				if ( port_letter == 'D' )
				{
					if ( port_num == 6 )
					{
						OCR0A = 0;
						TCCR0A |= (1 << COM0A1);
						// set fast PWM Mode
						TCCR0A |= (1 << WGM01) | (1 << WGM00);
						TCCR0B |= (1 << CS01);
					}

					if ( port_num == 5 )
					{
						OCR0B = 0;
						TCCR0A |= (1 << COM0B1);
						// set fast PWM Mode
						TCCR0A |= (1 << WGM01) | (1 << WGM00);
						TCCR0B |= (1 << CS01);
					}

					if ( port_num == 3 )
					{
						OCR2B = 0;

						TCCR2A |= (1 << COM2B1);
						// set fast PWM Mode
						TCCR2A |= (1 << WGM21) | (1 << WGM20);
						TCCR2B |= (1 << CS21);
					}

				}
			}
		}
		else
		{
			if ( (_port_type[i] == 2 || port_letter == 'A') && _port_m[i] > 0 && _port_m[i] < 255 )
			adc_check_flag = 1;

			if ( port_letter == 'D' )
			DDRD &= ~(1<<port_num);
			if ( port_letter == 'C' )
			DDRC &= ~(1<<port_num);
			if ( port_letter == 'B' )
			DDRB &= ~(1<<port_num);
		}

	}

	// Debouncing init
	TCCR1B |=   (1 << WGM12);         //Timer1 Mode 2: CTC
	TCCR1B |=   (1 << CS12);          //Divide by 256
	OCR1A    =   F_CPU / 512 * 10e-3; //Set CTC compare value to 5ms at 12,5 MHz AVR clock, with a prescaler of 256
	TIMSK1 =   (1 << OCIE1A);         //Enable T1 interrupt
	input_state    = ~PIND;      //no action on keypress during reset
	sei();

	//   
	// --- init_ip_arp_udp_tcp(mymac,myip,HTTPPORT);
        init_mac(mymac);
        client_ifconfig(myip,netmask);

	// Reading Gateway / Server IP address
	eeprom_read_block ((void *)_sip_addr, (const void *)&ee_gw_addr,4);
	if ( _sip_addr[0] == 255 )
	eeprom_read_block ((void *)_sip_addr, (const void *)&ee_sip_addr,4);
	// Reading script-name
	eeprom_read_block (_spt, &ee_spt, 15);
	// Reading mdid
	eeprom_read_block (_mdid, &ee_mdid, 5);
	if ( strlen(_mdid) == 0 )
	_mdid[0] = 255;

	if ( _sip_addr[0] != 255 && _sip_addr[0] > 0 )
	{
		LEDON;
		get_mac_with_arp(_sip_addr,TRANS_NUM_GWMAC,&arpresolver_result_callback);

		//      ARP- IP- 
		arp_timeout = 32000;
                while(get_mac_with_arp_wait() && arp_timeout > 0 )
		{
                        // to process the ARP reply we must call the packetloop
                        plen=enc28j60PacketReceive(BUFFER_SIZE, buf);
                        packetloop_arp_icmp_tcp(buf,plen);
			arp_timeout--;
                }

		arp_timeout = 0;

		get_mac_with_arp(_sip_addr,TRANS_NUM_WEBMAC,&arpresolver_result_callback);

		gw_set = 1;
		LEDOFF;

		//   SRAM   _sip_addr   Gateway IP,    Server IP
		eeprom_read_block ((void *)_sip_addr, (const void *)&ee_sip_addr,4);

	}

	temp_check_flag = eeprom_read_byte(&ee_temp_check);
	if ( temp_check_flag == 1 && ( _alarm_temp == 255 || _sip_addr[0] == 255 || _sip_addr[0] == 0 ) )
	temp_check_flag = 0;

	//srv_loop_timer = eeprom_read_byte(&ee_srv_loop) * SRV_LOOP_TIMEOUT;

	srv_loop_timer = eeprom_read_byte(&ee_srv_loop);
	if ( srv_loop_timer == 1 )
	srv_loop_timer = SRV_LOOP_TIMEOUT;
	else
	srv_loop_timer = 0;

        www_server_port(HTTPPORT);

	while(1)
	{
		plen=enc28j60PacketReceive(BUFFER_SIZE, buf);
		buf[BUFFER_SIZE]='\0';
		//--- dat_p=packetloop_icmp_tcp(buf,plen);
                dat_p=packetloop_arp_icmp_tcp(buf,plen);

		//    . URL.    ,   TCP-   . ( tuxgraphics TCP/IP)
		// .        ,    .   ?
		if ( send_eth_flag2 > -1  )
		{
			if ( (tcp_client_state == 0 || tcp_client_state >= 5) )
			{
	                        // to process the ARP reply we must call the packetloop
	                        plen=enc28j60PacketReceive(BUFFER_SIZE, buf);
	                        packetloop_arp_icmp_tcp(buf,plen);
				arp_timeout--;

				eeprom_read_block (_eth_addr, &ee_eth_cmd[cur_input], 4);
				//eeprom_read_block (_eth_addr, &ee_eth_cmd[atoi(gStrbuf)], 4);
				if ( _eth_addr[0] != 255 ) 
				{
					memset(temp, 0, sizeof(temp));
					//eeprom_read_block (temp, (void*)ee_eth_cmd[send_eth_flag] + 4, 12);
					eeprom_read_block (temp, (void*)ee_eth_cmd[cur_input] + 4, 12);

					get_mac_with_arp(_eth_addr,TRANS_NUM_WEBMAC,&arpresolver_result_callback2);

					arp_timeout = 32000;
			                while(get_mac_with_arp_wait() && arp_timeout > 0 )
					{
			                        // to process the ARP reply we must call the packetloop
			                        plen=enc28j60PacketReceive(BUFFER_SIZE, buf);
			                        packetloop_arp_icmp_tcp(buf,plen);
						arp_timeout--;
			                }

					if ( _mdid[0] != (char)255 && _mdid[0] != '\0' )
					{
						if ( strchr(temp, '?') )
						strcat(temp, "&");
						else
						strcat(temp, "?");
						strcat(temp, "mdid=");
						strcat(temp, _mdid);
					}

					sprintf(shost,"%d.%d.%d.%d",_eth_addr[0],_eth_addr[1],_eth_addr[2],_eth_addr[3]);
					LEDON;
					client_browse_url(PSTR("/"),temp,shost,&browserresult_callback_empty, _eth_addr, www_gwmac2);
					LEDOFF;

				}

				//send_eth_flag = -1;
				send_eth_flag2 = -1;
			}

		}

		//   
		for ( i = 0; i < IO_SIZE; i++ )
		{
			port_letter = aio[i][0];
			port_num = atoi(&aio[i][1]);
			//cur_input = i;

			//   ...

			//if ( _port_type[i] == 0 )
			if ( _port_type[i] == 0 || ( _port_type[i] == 2 || port_letter == 'A') )
			{
				//  ,    
				if ( srv_timeout_act[i] == 1 )
				{
					srv_timeout_act[i] = 0;
					memset(temp, 0, sizeof(temp));
					eeprom_read_block (temp, &ee_cmd[i], 23);
					if ( temp[0] != (char)255 && strlen(temp) > 0 ) 
					port_execute(temp, i, 0);
					//      . ,       EEPROM URL
					//  2 -   default gateway,     
					tcp_client_state = 0;
					send_eth_flag2 = 2;
				}
			}

			if ( _port_type[i] == 0 || ( (_port_type[i] == 2 || port_letter == 'A') && port_act[i] == 1 ) )
			{

				char my_mask = '\0';
				char my_mask2 = '\0';

				if ( _port_type[i] == 0 )
				{
					//   .
					if (port_letter == 'D')
					{
						my_mask = get_key_press(1 << port_num);
						my_mask2 = get_key_release(1 << port_num);
					}
					else if (port_letter == 'C')
					{
						my_mask = get_key_press2(1 << port_num);
						my_mask2 = get_key_release2(1 << port_num);
					}
					else
					{
						my_mask = get_key_press3(1 << port_num);
						my_mask2 = get_key_release3(1 << port_num);
					}
				}

				if ( port_act[i] == 1 || ( ( (_port_m[i] == 0 || _port_m[i] == 255) && my_mask ) || ( _port_m[i] == 1 && (my_mask || my_mask2) ) || ( _port_m[i] == 2 && my_mask2 ) ) || srv_timeout[i] > 150 )
				{

					//port_act[i] = 0;
					//port_cnt[i]++;

					//cur_input = i;

					//    
					if ( _sip_addr[0] != 255 && enc28j60linkup() )
					{
						//    TCP-   ,   -  ,   
						if ( (tcp_client_state == 0 || tcp_client_state >= 5) )
						{
							port_act[i] = 0;
							port_cnt[i]++;

							char pnum[3];
							char adc_val[5];
							itoa(i, pnum, 10);

							sprintf(shost,"%d.%d.%d.%d",_sip_addr[0],_sip_addr[1],_sip_addr[2],_sip_addr[3]);
							strcpy(temp, _spt);
							strcat(temp, "?pt=");
							strcat(temp, pnum);
							if ( my_mask2 || (_port_type[i] == 0 && port_misc_cur[i] == 1 ) )
							{
								strcat(temp, "&m=1");
								port_misc_cur[i] = 0;
							}
							if (_port_type[i] == 2 || port_letter == 'A' )
							{
								strcat(temp, "&v=");
								itoa(port_misc_cur[i], adc_val, 10);
								strcat(temp, adc_val);
								strcat(temp, "&dir=");
								if ( port_misc_cur[i] > _port_misc[i] )
								strcat(temp, "1");
								else
								strcat(temp, "0");
							}

							if ( _mdid[0] != (char)255 )
							{
								strcat(temp, "&mdid=");
								strcat(temp, _mdid);
							}

							cur_input = i;

							LEDON;
							client_browse_url(PSTR("/"),temp,shost,&browserresult_callback, _sip_addr, gwmac);
							LEDOFF;

							srv_timeout[i] = 150;

							if ( srv_loop_timer > 0 )
							srv_loop_timer = 1000;
						}
						else if ( srv_timeout[i] == 0 )
						{
							srv_timeout[i] = 250;
							//    (P  R) -         ,
							//   port_misc_cur (    )   .
							if ( my_mask2 )
						        port_misc_cur[i] = 1;
						}

						//send_eth_flag = cur_input;

					}
					//    
					else if ( _sip_addr[0] == 255 || !enc28j60linkup() )
					{
						//      . ,       EEPROM URL
						send_eth_flag2 = 1;
						memset(temp, 0, sizeof(temp));
						eeprom_read_block (temp, &ee_cmd[i], 23);
						//  ,   EEPROM   
						if ( temp[0] != (char)255 && strlen(temp) > 0 ) 
						port_execute(temp, i, 0);

						//send_eth_flag = cur_input;
					}

					//send_eth_flag = cur_input;

				}
			}

			if ( srv_timeout_act[i] == 2 )
			{
				srv_timeout_act[i] = 0;

				char temp2[23];
				memset(temp, 0, sizeof(temp));
				eeprom_read_block(temp2, (void*)ee_cmd[i] + cmd_delay_p[i], 23 - cmd_delay_p[i]);
				port_execute(temp2, i, cmd_delay_p[i]);
			}

		}

		if ( adc_check_flag == 1 )
		{
			if ( adc_timer > 1000 )
			{
				for ( i = 0; i < IO_SIZE; i++ )
				{
					port_letter = aio[i][0];
					port_num = atoi(&aio[i][1]);

					if (_port_type[i] == 2 || port_letter == 'A' )
					{
						if ( _port_misc[i] < 1024 && _port_m[i] != 0 )
						{
							ADMUX = (1 << REFS0) + port_num; 
							ADCSRA|=(1<<ADSC);
							while (bit_is_set(ADCSRA, ADSC));
							uint16_t my_val;
							uint8_t my_val_low;
							my_val_low = ADCL;
							my_val = (ADCH<<8)|my_val_low;
	
							if ( (_port_m[i] == 1 && my_val > _port_misc[i] && port_misc_cur[i] < _port_misc[i] ) || (_port_m[i] == 2 && my_val < _port_misc[i] && port_misc_cur[i] > _port_misc[i] ) || ( _port_m[i] == 3 && ( (my_val > _port_misc[i] && port_misc_cur[i] < _port_misc[i]) || (my_val < _port_misc[i] && port_misc_cur[i] > _port_misc[i] ) ) ) )
							port_act[i] = 1;

							port_misc_cur[i] = my_val;
						}
					}
				}

				adc_timer = 0;
			}
		}

		if ( temp_check_flag > 0 )
		{
			if ( temp_timer > 12000 )
			{
				uint8_t cur_temp = readchiptemp();
				char temp2[4];

				if ( cur_temp >= _alarm_temp )
				{
					sprintf(shost,"%d.%d.%d.%d",_sip_addr[0],_sip_addr[1],_sip_addr[2],_sip_addr[3]);
					strcpy(temp, _spt);
					strcat(temp, "?at=");
					sprintf(temp2,"%d", readchiptemp());
					strcat(temp, temp2);
					LEDON;
					client_browse_url(PSTR("/"),temp,shost,&browserresult_callback, _sip_addr, gwmac);
					LEDOFF;
				}
		
				temp_timer = 0;
			}
		}

		if ( srv_loop_timer == 1 )
		{

			if (tcp_client_state == 0 || tcp_client_state >= 5)
			{
				sprintf(shost,"%d.%d.%d.%d",_sip_addr[0],_sip_addr[1],_sip_addr[2],_sip_addr[3]);

				for ( i = 0; i < 100; i++ )
				{
					if ( buf[i+TCP_CHECKSUM_L_P+3+plen_start] == 42 )
					buf[700+i] = '\0';
					else
					//break;

					buf[700+i] = buf[i+TCP_CHECKSUM_L_P+3+plen_start];

				}

				LEDON;
				client_browse_url(PSTR("/"),(char*)&(buf[700]),shost,&browserresult_callback, _sip_addr, gwmac);
				LEDOFF;

				srv_loop_timer = SRV_LOOP_TIMEOUT;

			}
			else
			srv_loop_timer = 1000;
		}

		if ( srv_loop_timer != 2 )
		{
	                if(dat_p==0)
			{
	                        // check if udp otherwise continue
				uint8_t broadcastmac[6] = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF}; 

				typedef struct  
				{
					uint8_t header;
					uint8_t msg_id;
					uint8_t cmd;
					uint8_t data[14];
				} udp_data_struct;
	
				udp_data_struct* udp_data = (udp_data_struct*) (buf + UDP_DATA_P);
				#define check_new_UDP_message() ((plen>0) && (buf[IP_PROTO_P]==IP_PROTO_UDP_V))

			        if ( check_new_UDP_message() && (udp_data->header == 0xAA) )
				{
					send_udp_prepare(buf,52000,&buf[IP_SRC_P],42000, broadcastmac);
					buf[UDP_DATA_P] = 0xAA;

					if (udp_data->cmd == 12)
					{
						for ( i = 0; i < 4; i++ )
						buf[UDP_DATA_P + i + 1] = myip[i];
						send_udp_transmit(buf,5);
					}

					else if (udp_data->cmd == 4 )
					{
						uint8_t ip_check = 0;

						for ( i = 0; i < 4; i++ )
						{
							if ( udp_data->data[5 + i] != myip[i] )
							ip_check = 1;
						}

						if ( ip_check == 0 )
						{
							if ( strncmp(password,(void *)udp_data->data,5) == 0 )
							{
								eeprom_write_block ((void *)udp_data->data+9, (void *)ee_ip_addr, 4);
								buf[UDP_DATA_P + 1] = 0x01;
							}
							else
							buf[UDP_DATA_P + 1] = 0x02;

							send_udp_transmit(buf,2);

							if ( strncmp(password,(void *)udp_data->data,5) == 0 )
							RESET();
						}
					}

				}

	                        goto UDP;
			}

			if (strncmp("/ ",(char *)&(buf[dat_p+4]),2)==0){
				plen=http200ok();
				plen=fill_tcp_data_p(buf,plen,PSTR("<p>Usage: http://host_or_ip/password</p>\n"));
				goto SENDTCP;
			}

			cmd=analyse_get_url((char *)&(buf[dat_p+4]));
	
			if (cmd==-1)
			{
				plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1>"));
				goto SENDTCP;
			}
		}

		//    
		uint8_t ch_port_flag = 0;
		//  
		uint8_t mode = 0;
		uint8_t start_loop = 0;
		uint8_t end_loop = 0;
		// 
		char srv_cmd[33];
		//     
		memset(srv_cmd, 0, 33);
		if (find_key_val((char *)&(buf[dat_p+4]),srv_cmd,33,"cmd"))
		{
			if ( strcmp(srv_cmd, "get") == 0 )
			mode = 1;
			if ( strlen(srv_cmd) > 0 )
			port_execute(srv_cmd, 100, 0);
		}

		if ( srv_loop_timer == 2 && tcp_client_state > 0 && tcp_client_state < 5 )
		srv_loop_timer = 500;

		if ( find_key_val((char *)&(buf[dat_p+4]),gStrbuf,25,"pt") || strcmp(srv_cmd, "all") == 0 || srv_loop_timer == 2 )
		{
			//    
			if ( strcmp(srv_cmd, "all") == 0 || srv_loop_timer == 2 )
			{
				end_loop = IO_SIZE;
				mode = 1;
			}
			else
			{
				start_loop = atoi(gStrbuf);
				end_loop = start_loop + 1;

				if (find_key_val((char *)&(buf[dat_p+4]),temp,25,"cnt"))
				port_cnt[start_loop] = atoi(temp);
			}


			//   PWM
			if (find_key_val((char *)&(buf[dat_p+4]),temp,4,"pwm"))
			{
				port_letter = aio[atoi(gStrbuf)][0];
				port_num = atoi(&aio[atoi(gStrbuf)][1]);

				if ( port_letter == 'D' )
				{
					if ( port_num == 6 )
					OCR0A = atoi(temp);
					else if ( port_num == 5 )
					OCR0B = atoi(temp);
					else if ( port_num == 3 )
					OCR2B = atoi(temp);
				}
			}

			eeprom_read_block (_port_type, &ee_port_type, IO_SIZE);

		        plen=http200ok();
			plen_start = plen;

			if ( srv_loop_timer == 2 )
			{
				plen=fill_tcp_data(buf,plen,_spt);

				plen=fill_tcp_data_p(buf,plen,PSTR("?"));
				if ( _mdid[0] != (char)255 )
				{
					plen=fill_tcp_data_p(buf,plen,PSTR("mdid="));
					plen=fill_tcp_data(buf,plen,_mdid);
				}
				plen=fill_tcp_data_p(buf,plen,PSTR("&all="));
			}


			for ( i = start_loop; i < end_loop; i++ )
			{

				port_letter = aio[i][0];
				port_num = atoi(&aio[i][1]);
				//port_letter = aio[atoi(gStrbuf)][0];
				//port_num = atoi(&aio[atoi(gStrbuf)][1]);
			
				if ( port_letter == 'D' )
				{
					avr_port = DDRD;
					avr_port_port = PORTD;
					avr_port_pin = PIND;
				}
				else if ( port_letter == 'C' )
				{
					avr_port = DDRC;
					avr_port_port = PORTC;
					avr_port_pin = PINC;
				}
				else if ( port_letter == 'B' )
				{
					avr_port = DDRB;
					avr_port_port = PORTB;
					avr_port_pin = PINB;
				}

				if ( mode == 0 )
				{

					plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/"));
					plen=fill_tcp_data(buf,plen,password);
					plen=fill_tcp_data_p(buf,plen,PSTR(">Back</a><br>"));
	
					if ( port_letter == 'A' )
					{
						plen=fill_tcp_data_p(buf,plen,PSTR("Port: A"));
						plen=fill_tcp_data(buf,plen,&aio[i][1]);
					}
					else
					{
						plen=fill_tcp_data_p(buf,plen,PSTR("Port: P"));
						plen=fill_tcp_data(buf,plen,gStrbuf);
					}
				}

				//    
				if ( port_num < IO_SIZE && ( _port_type[i] != 255 || ( port_letter == 'A' && ( port_num == 6 || port_num == 7 ) ) ) )
				{

					//    
					if ( _port_type[i] == 2 || port_letter == 'A' )
					{
						if ( mode == 0 )
						plen=fill_tcp_data_p(buf,plen,PSTR("/"));

						//if ( port_letter == 'A' )
						//ADMUX = (1 << REFS0) + 6; 
						//else
						ADMUX = (1 << REFS0) + port_num; 
						//ADMUX |= (1 << ADLAR); 
						ADCSRA|=(1<<ADSC);
						while (bit_is_set(ADCSRA, ADSC));
						//_delay_loop_1(0); // 60us
						//_delay_loop_1(0); // 60us
						uint16_t my_val;
						uint8_t my_val_low;
						my_val_low = ADCL;
						my_val = (ADCH<<8)|my_val_low;
						//snprintf_P(temp,sizeof(temp),PSTR("%d"),my_val);
						itoa (my_val, temp, 10);
						plen=fill_tcp_data(buf,plen, temp);
					}
					//    
					else if ( bit_is_set(avr_port, port_num) )
					{
						//   PWM
						if (_port_m[i] == 1 )
						{

							if ( mode == 0 )
							{
								plen=fill_tcp_data_p(buf,plen,PSTR("<br><form style=display:inline action=/"));
								plen=fill_tcp_data(buf,plen,password);
								plen=fill_tcp_data_p(buf,plen,PSTR("/>"));
								plen=fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=pt value="));
								plen=fill_tcp_data(buf,plen,gStrbuf);
								plen=fill_tcp_data_p(buf,plen,PSTR(">"));
								plen=fill_tcp_data_p(buf,plen,PSTR("PWM (0-255): <input type=text name=pwm size=3 value="));
							}

							if ( port_letter == 'D' )
							{
								if ( port_num == 5 )
								itoa(OCR0B, temp, 10);
								else if ( port_num == 6 )
								itoa(OCR0A, temp, 10);
								else if ( port_num == 3 )
								itoa(OCR2B, temp, 10);
							}
							plen=fill_tcp_data(buf,plen,temp);
						
							if ( mode == 0 )
							plen=fill_tcp_data_p(buf,plen,PSTR("> <input type=submit value=change></form>"));

						}
						//   SWITCH
						else
						{
							if ( mode == 0 )
							plen=fill_tcp_data_p(buf,plen,PSTR("/"));
		
							if ( bit_is_set(avr_port_port, port_num) )
							plen=fill_tcp_data_p(buf,plen,PSTR("ON"));
							else
							plen=fill_tcp_data_p(buf,plen,PSTR("OFF"));

							if ( mode == 0 )
							{
								plen=fill_tcp_data_p(buf,plen,PSTR("<br>"));

								plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/"));
								plen=fill_tcp_data(buf,plen,password);
								plen=fill_tcp_data_p(buf,plen,PSTR("/?pt="));
								plen=fill_tcp_data(buf,plen,gStrbuf);
								plen=fill_tcp_data_p(buf,plen,PSTR("&cmd="));
								plen=fill_tcp_data(buf,plen,gStrbuf);
								plen=fill_tcp_data_p(buf,plen,PSTR(":1>ON</a> "));

								plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/"));
								plen=fill_tcp_data(buf,plen,password);
								plen=fill_tcp_data_p(buf,plen,PSTR("/?pt="));
								plen=fill_tcp_data(buf,plen,gStrbuf);
								plen=fill_tcp_data_p(buf,plen,PSTR("&cmd="));
								plen=fill_tcp_data(buf,plen,gStrbuf);
								plen=fill_tcp_data_p(buf,plen,PSTR(":0>OFF</a>"));

								plen=fill_tcp_data_p(buf,plen,PSTR("<br>"));
							}
						}
					}
					else if ( _port_type[i] == 3 )
					{
						dht11_temperature = 0;
						dht11_humidity = 0;
						dht_type = _port_m[i];

						dht_gettemperaturehumidity(&dht11_temperature, &dht11_humidity, _port_m[i]);
						{
							itoa(dht11_temperature, temp, 10);

							if ( mode == 0 )
							plen=fill_tcp_data_p(buf,plen,PSTR("<br>"));
							plen=fill_tcp_data_p(buf,plen,PSTR("temp:"));
							plen=fill_tcp_data(buf,plen,temp);

							itoa(dht11_humidity, temp, 10);
	
							if ( mode == 0 )
							plen=fill_tcp_data_p(buf,plen,PSTR("<br>"));
							else
							plen=fill_tcp_data_p(buf,plen,PSTR("/"));
							plen=fill_tcp_data_p(buf,plen,PSTR("hum:"));
							plen=fill_tcp_data(buf,plen,temp);
						}
	
					}
					//    
					else
					{
						if ( mode == 0 )
						plen=fill_tcp_data_p(buf,plen,PSTR("/"));

						if ( bit_is_clear(avr_port_pin, port_num) )
						plen=fill_tcp_data_p(buf,plen,PSTR("ON"));
						else
						plen=fill_tcp_data_p(buf,plen,PSTR("OFF"));
	
						plen=fill_tcp_data_p(buf,plen,PSTR("/"));
						utoa(port_cnt[i], temp, 10);
						plen=fill_tcp_data(buf,plen,temp);

						//if ( mode != 0 )
						//plen=fill_tcp_data_p(buf,plen,PSTR("<br>"));
					}
				}

				if ( mode == 0)
				{
					plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/"));
					plen=fill_tcp_data(buf,plen,password);
					plen=fill_tcp_data_p(buf,plen,PSTR("/>"));
					plen=fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=pn value="));
					plen=fill_tcp_data(buf,plen,gStrbuf);
					plen=fill_tcp_data_p(buf,plen,PSTR(">"));

					if ( port_letter != 'A' && _preset != 1 )
					{
						plen=fill_tcp_data_p(buf,plen,PSTR("Type: <select name=pty>"));

						if ( _port_type[i] == 255 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option selected value=255>NC</option><option value=0>In</option><option value=1>Out</option><option value=3>DSen</option>"));
						else if ( bit_is_set(avr_port, port_num) && _port_type[i] == 1 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=255>NC</option><option value=0>In</option><option value=1 selected>Out</option><option value=3>DSen</option>"));
						else if ( bit_is_clear(avr_port, port_num) && _port_type[i] == 0 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=255>NC</option><option value=0 selected>In</option><option value=1>Out</option><option value=3>DSen</option>"));
						else if ( _port_type[i] == 3 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=255>NC</option><option value=0>In</option><option value=1>Out</option><option value=3 selected>DSen</option>"));
						else
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=255>NC</option><option value=0>In</option><option value=1>Out</option><option value=3>DSen</option>"));

						if ( port_letter == 'C' )
						{
							if ( _port_type[i] == 2 )
							plen=fill_tcp_data_p(buf,plen,PSTR("<option value=2 selected>ADC</option>"));
							else
							plen=fill_tcp_data_p(buf,plen,PSTR("<option value=2>ADC</option>"));
						}

						plen=fill_tcp_data_p(buf,plen,PSTR("</select>"));
					}
					else if ( port_letter != 'A' && _preset == 1 )
					{
						plen=fill_tcp_data_p(buf,plen,PSTR("Type: "));
						if ( bit_is_clear(avr_port, port_num) && _port_type[i] == 0 )
						plen=fill_tcp_data_p(buf,plen,PSTR("In"));
						else if ( bit_is_set(avr_port, port_num) && _port_type[i] == 1 )
						plen=fill_tcp_data_p(buf,plen,PSTR("Out"));
						else
						plen=fill_tcp_data_p(buf,plen,PSTR("NC"));
					}

					if ( _port_type[i] == 2 || port_letter == 'A' )
					{
						if ( _port_type[i] == 2 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<br>"));

						plen=fill_tcp_data_p(buf,plen,PSTR("Mode: <select name=m>"));
						if (_port_m[i] == 1 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0>Norm</option><option selected value=1>></option><option value=2><</option><option value=3><></option>"));
						else if (_port_m[i] == 2 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0>Norm</option><option value=1>></option><option selected value=2><</option><option value=3><></option>"));
						else if (_port_m[i] == 3 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0>Norm</option><option value=1>></option><option value=2><</option><option selected value=3><></option>"));
						else
						plen=fill_tcp_data_p(buf,plen,PSTR("<option selected value=0>Norm</option><option value=1>></option><option value=2><</option><option value=3><></option>"));
						plen=fill_tcp_data_p(buf,plen,PSTR("</select>"));

						plen=fill_tcp_data_p(buf,plen,PSTR("<br>Val: <input type=text name=misc size=4 maxlength=4 value='"));
						if ( _port_misc[i] != 65535 )
						{
							sprintf(temp,"%u",_port_misc[i]);
							plen=fill_tcp_data(buf,plen,temp);
						}
						plen=fill_tcp_data_p(buf,plen,PSTR("'>"));

					}

					if ( ( bit_is_clear(avr_port, port_num) && _port_type[i] == 0 ) || _port_type[i] == 2 || port_letter == 'A' )
					{
						plen=fill_tcp_data_p(buf,plen,PSTR("<br>Action: <input type=text name=ecmd size=20 maxlength=23 value=\""));
						eeprom_read_block (temp, &ee_cmd[i], 23);
						if ( temp[0] != (char)255 && strlen(temp) > 0 ) 
						plen=fill_tcp_data(buf,plen, temp);
						plen=fill_tcp_data_p(buf,plen,PSTR("\"><br>"));

						plen=fill_tcp_data_p(buf,plen,PSTR("NetAction: <input type=text maxlength=28 size=30 name=eth value=\""));

						eeprom_read_block (_eth_addr, &ee_eth_cmd[i], 4);
						if ( _eth_addr[0] != 255 )
						{
							snprintf_P(temp,sizeof(temp),PSTR("%d.%d.%d.%d"),(int)_eth_addr[0],(int)_eth_addr[1],(int)_eth_addr[2],(int)_eth_addr[3]);
							plen=fill_tcp_data(buf,plen, temp);
							plen=fill_tcp_data_p(buf,plen,PSTR("/"));

							memset(temp, 0, sizeof(temp));
							eeprom_read_block (temp, (void*)ee_eth_cmd[i] + 4, 12);
							plen=fill_tcp_data(buf,plen, temp);
						}
						plen=fill_tcp_data_p(buf,plen,PSTR("\">"));
					}

					if ( bit_is_clear(avr_port, port_num) && _port_type[i] == 0 )
					{
						plen=fill_tcp_data_p(buf,plen,PSTR("<br>Mode: <select name=m>"));
						if (_port_m[i] == 1 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0>P</option><option selected value=1>P&R</option><option value=2>R</option>"));
						else if (_port_m[i] == 2 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0>P</option><option value=1>P&R</option><option selected value=2>R</option>"));
						else
						plen=fill_tcp_data_p(buf,plen,PSTR("<option selected value=0>P</option><option value=1>P&R</option><option value=2>R</option>"));
						plen=fill_tcp_data_p(buf,plen,PSTR("</select>"));
					}
					else if ( bit_is_set(avr_port, port_num) && _port_type[i] == 1 )
					{
						//     
						plen=fill_tcp_data_p(buf,plen,PSTR("<br>Default state: <select name=d>"));
						if (_port_d[i] == 1 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0>0</option><option selected value=1>1</option>"));
						else
						plen=fill_tcp_data_p(buf,plen,PSTR("<option selected value=0>0</option><option value=1>1</option>"));
						plen=fill_tcp_data_p(buf,plen,PSTR("</select>"));

						//  : /
						if ( _preset != 1 && port_letter == 'D' && ( port_num == 3 || port_num == 5 || port_num == 6 ) )
						{
							plen=fill_tcp_data_p(buf,plen,PSTR("<br>Mode: <select name=m>"));
							if (_port_m[i] == 1 )
							plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0>SW</option><option selected value=1>PWM</option>"));
							else
							plen=fill_tcp_data_p(buf,plen,PSTR("<option selected value=0>SW</option><option value=1>PWM</option>"));
							plen=fill_tcp_data_p(buf,plen,PSTR("</select>"));
						}
					}
					else if ( _port_type[i] == 3 )
					{
						plen=fill_tcp_data_p(buf,plen,PSTR("<br>Sensor: <select name=m>"));
						if (_port_m[i] == 2 )
						plen=fill_tcp_data_p(buf,plen,PSTR("<option value=1>DHT11</option><option selected value=2>DHT22</option>"));
						else
						plen=fill_tcp_data_p(buf,plen,PSTR("<option selected value=1>DHT11</option><option value=2>DHT22</option>"));
						plen=fill_tcp_data_p(buf,plen,PSTR("</select>"));
					}

					plen=fill_tcp_data_p(buf,plen,PSTR("<br><input type=submit value=Save></form>"));
				}

				if ( mode == 1 && i < end_loop - 1 )
				plen=fill_tcp_data_p(buf,plen,PSTR(";"));
			}

			if ( srv_loop_timer == 2 )
			{
				plen=fill_tcp_data_p(buf,plen,PSTR("*"));
				srv_loop_timer = 1;

			
				goto UDP;
			}

		}
		else if (find_key_val((char *)&(buf[dat_p+4]),gStrbuf,25,"pn"))
		{
			memset(temp, 0, sizeof(temp));

			if (find_key_val((char *)&(buf[dat_p+4]),temp,25,"pty"))
			{
				//    .    
				if ( _port_type[atoi(gStrbuf)] != atoi(temp) )
				{
					ch_port_flag = 1;
					_port_m[atoi(gStrbuf)] = 255;
					eeprom_write_block (&_port_m[atoi(gStrbuf)], &ee_port_m[atoi(gStrbuf)], 1);
				}

				if ( strcmp(temp,"NC") )
				{
					_port_type[atoi(gStrbuf)] = atoi(temp);
					eeprom_write_block (&_port_type[atoi(gStrbuf)], &ee_port_type[atoi(gStrbuf)], 1);
					reset_flag = 1;
				}
			}

			//      ,     
			if ( ch_port_flag == 0 )
			{
				if (find_key_val2((char *)&(buf[dat_p+4]),temp,42,"ecmd"))
				{
					//if ( strlen(temp) > 0 )
					{
						urldecode(temp);
						eeprom_write_block (&temp, &ee_cmd[atoi(gStrbuf)], 23);
						reset_flag = 1;
					}
				}

				if (find_key_val2((char *)&(buf[dat_p+4]),temp,42,"eth"))
				{
					//if ( strlen(temp) > 0 )
					{
						urldecode(temp);

						char urlpar[12];
						uint8_t urlpar_flag = 0;

						if ( temp[0] == '\0' )
						_eth_addr[0] = 255;
						else
						{
							memset(urlpar, 0, sizeof(urlpar));

							for ( i = 0; i < strlen(temp); i++ )
							{
								if ( urlpar_flag > 0 )
								{
									urlpar[urlpar_flag-1] = temp[i];
									urlpar_flag++;
								}
								else
								{
									if ( temp[i] == '/' )
									urlpar_flag++;
								}
							}

							decode_ip(temp, _eth_addr);
						}

						eeprom_write_block (_eth_addr, &ee_eth_cmd[atoi(gStrbuf)], 4);

						eeprom_write_block (urlpar, (void*)ee_eth_cmd[atoi(gStrbuf)]+4, 12);
						reset_flag = 1;
					}
				}

				if (find_key_val((char *)&(buf[dat_p+4]),temp,25,"m"))
				{
					_port_m[atoi(gStrbuf)] = atoi(temp);
					eeprom_write_block (&_port_m[atoi(gStrbuf)], &ee_port_m[atoi(gStrbuf)], 1);
				}

				if (find_key_val((char *)&(buf[dat_p+4]),temp,25,"d"))
				{
					_port_d[atoi(gStrbuf)] = atoi(temp);
					eeprom_write_block (&_port_d[atoi(gStrbuf)], &ee_port_d[atoi(gStrbuf)], 1);
				}

				if (find_key_val((char *)&(buf[dat_p+4]),temp,25,"misc"))
				{
					_port_misc[atoi(gStrbuf)] = atoi(temp);
					eeprom_write_block (&_port_misc[atoi(gStrbuf)], &ee_port_misc[atoi(gStrbuf)], 2);
				}

			}


		        plen=http200ok();
			plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/"));
			plen=fill_tcp_data(buf,plen,password);
			plen=fill_tcp_data_p(buf,plen,PSTR("/>Back</a>"));

			if ( reset_flag == 1 )
			plen=fill_tcp_data_p(buf,plen,PSTR("<br>Device is restared."));

		}
		else if (find_key_val((char *)&(buf[dat_p+4]),gStrbuf,2,"cf"))
		{
			uint8_t temp_check_flag2 = 0;
			uint8_t cf_mode = atoi(gStrbuf);

			//if ( gStrbuf[0] == '1' )
			{
				if (find_key_val((char *)&(buf[dat_p+4]),gStrbuf,25,"eip"))
				{
					decode_ip(gStrbuf, _ip_addr);
					eeprom_write_block (_ip_addr, (void *) ee_ip_addr, 4);
					reset_flag = 1;
				}

				if (find_key_val2((char *)&(buf[dat_p+4]),gStrbuf,25,"sip"))
				{
					if ( !strcmp(gStrbuf,"") )
					{
						_sip_addr[0] = 255;
						_sip_addr[1] = 255;
						_sip_addr[2] = 255;
						_sip_addr[3] = 255;
					}
					else
					decode_ip(gStrbuf, _sip_addr);
					eeprom_write_block (_sip_addr, &ee_sip_addr, 4);
					reset_flag = 1;
				}

				// Gateway IP
				if (find_key_val2((char *)&(buf[dat_p+4]),gStrbuf,25,"gw"))
				{
					if ( !strcmp(gStrbuf,"") )
					{
						_sip_addr[0] = 255;
						_sip_addr[1] = 255;
						_sip_addr[2] = 255;
						_sip_addr[3] = 255;
					}
					else
					decode_ip(gStrbuf, _sip_addr);
					eeprom_write_block (_sip_addr, &ee_gw_addr, 4);
					reset_flag = 1;
				}


				if (find_key_val2((char *)&(buf[dat_p+4]),gStrbuf,25,"sct"))
				{
					eeprom_write_block (gStrbuf, &ee_spt, 15);
					reset_flag = 1;
				}

				if (find_key_val((char *)&(buf[dat_p+4]),gStrbuf,25,"pwd"))
				{
					eeprom_write_block (gStrbuf, &ee_pwd, 6);
					strcpy(password, gStrbuf);
					reset_flag = 1;
				}

				if (find_key_val2((char *)&(buf[dat_p+4]),gStrbuf,25,"at"))
				{
					eeprom_write_byte((uint8_t*)&ee_alarm_temp, atoi(gStrbuf));
					_alarm_temp=atoi(gStrbuf);
					reset_flag = 1;
				}

				if (find_key_val2((char *)&(buf[dat_p+4]),gStrbuf,25,"mdid"))
				{
					eeprom_write_block (gStrbuf, &ee_mdid, 5);
					reset_flag = 1;

					if (find_key_val((char *)&(buf[dat_p+4]),gStrbuf,25,"sl"))
					eeprom_write_byte((uint8_t*)&ee_srv_loop, 1);
					else
					eeprom_write_byte((uint8_t*)&ee_srv_loop, 0);

				}


				if (find_key_val((char *)&(buf[dat_p+4]),gStrbuf,25,"tc"))
				{
					eeprom_write_byte((uint8_t*)&ee_temp_check, 1);
					temp_check_flag = 1;
					temp_check_flag2 = 1;
					reset_flag = 1;
				}

				if (find_key_val2((char *)&(buf[dat_p+4]),gStrbuf,25,"pr"))
				{
					//eeprom_write_block(atoi(gStrbuf), &ee_preset, 1);
					if ( _preset != atoi(gStrbuf) )
					{
						eeprom_write_byte((uint8_t*)&ee_preset, atoi(gStrbuf));

						if ( atoi(gStrbuf) == 1 )
						{
							for ( i = 0; i < IO_SIZE; i++ )
							{
								port_letter = aio[i][0];
								//port_num = atoi(&aio[i][1]);
								if ( i < 7 )
								_port_type[i] = 0;
								else if ( port_letter == 'A' )
								_port_type[i] = 255;
								else
								_port_type[i] = 1;

								if ( port_letter != 'A' )
								_port_m[i] = 255;
							}

							eeprom_write_block (&_port_m, &ee_port_m, sizeof(_port_m));
							eeprom_write_block (&_port_type, &ee_port_type, sizeof(_port_type));
							_preset = 1;
							reset_flag = 1;
						}
						else
						_preset = 0;
					}

					if ( temp_check_flag2 == 0 )
					eeprom_write_byte((uint8_t*)&ee_temp_check, 0);

					reset_flag = 1;
				}

			        plen=http200ok();

				eeprom_read_block ((void *)_ip_addr, (const void *)ee_ip_addr,4);
				eeprom_read_block ((void *)_sip_addr, (const void *)&ee_sip_addr,4);
				eeprom_read_block (_spt, &ee_spt,15);
				eeprom_read_block (_mdid, &ee_mdid,5);
				//eeprom_read_block (password, &ee_pwd,6);

				if ( _ip_addr[0] == 255 )
				{
					for ( i = 0; i < 4; i++ )
					_ip_addr[i] = myip[i];
				}

				plen=fill_tcp_data_p(buf,plen,PSTR("<a href=/"));
				plen=fill_tcp_data(buf,plen,password);
				plen=fill_tcp_data_p(buf,plen,PSTR(">Back</a> | <a href=/"));
				plen=fill_tcp_data(buf,plen,password);
				if ( cf_mode == 1 )
				plen=fill_tcp_data_p(buf,plen,PSTR("/?cf=2>Megad-ID</a><br>"));
				else if ( cf_mode == 2 )
				plen=fill_tcp_data_p(buf,plen,PSTR("/?cf=1>Config</a><br>"));


				plen=fill_tcp_data_p(buf,plen,PSTR("<form action=/"));
				plen=fill_tcp_data(buf,plen,password);
				plen=fill_tcp_data_p(buf,plen,PSTR("/>"));

				plen=fill_tcp_data_p(buf,plen,PSTR("<input type=hidden name=cf value="));

				if ( cf_mode == 1 )
				{

					plen=fill_tcp_data_p(buf,plen,PSTR("1>"));

					plen=fill_tcp_data_p(buf,plen,PSTR("IP: <input type=text name=eip value="));
					snprintf_P(temp,sizeof(temp),PSTR("%d.%d.%d.%d"),(int)_ip_addr[0],(int)_ip_addr[1],(int)_ip_addr[2],(int)_ip_addr[3]);
					plen=fill_tcp_data(buf,plen, temp);
					plen=fill_tcp_data_p(buf,plen,PSTR("><br>"));

					plen=fill_tcp_data_p(buf,plen,PSTR("Pwd: <input type=text name=pwd maxlength=3 value=\""));
					plen=fill_tcp_data(buf,plen,password);
					plen=fill_tcp_data_p(buf,plen,PSTR("\"><br>"));

					plen=fill_tcp_data_p(buf,plen,PSTR("GW: <input type=text name=gw value="));
					eeprom_read_block ((void *)_sip_addr, (const void *)&ee_gw_addr,4);
					snprintf_P(temp,sizeof(temp),PSTR("%d.%d.%d.%d"),(int)_sip_addr[0],(int)_sip_addr[1],(int)_sip_addr[2],(int)_sip_addr[3]);
					plen=fill_tcp_data(buf,plen, temp);
					plen=fill_tcp_data_p(buf,plen,PSTR("><br>"));

					plen=fill_tcp_data_p(buf,plen,PSTR("Server: <input type=text name=sip value="));
					eeprom_read_block ((void *)_sip_addr, (const void *)&ee_sip_addr,4);
					snprintf_P(temp,sizeof(temp),PSTR("%d.%d.%d.%d"),(int)_sip_addr[0],(int)_sip_addr[1],(int)_sip_addr[2],(int)_sip_addr[3]);
					plen=fill_tcp_data(buf,plen, temp);
					plen=fill_tcp_data_p(buf,plen,PSTR("><br>"));

					plen=fill_tcp_data_p(buf,plen,PSTR("Script: <input type=text name=sct maxlength=15 value=\""));
					//if ( _spt[0] != '' && strlen(_spt) > 0 ) 
					if ( _spt[0] != (char)255 && strlen(_spt) > 0 )
					plen=fill_tcp_data(buf,plen,_spt);
					plen=fill_tcp_data_p(buf,plen,PSTR("\">"));

					plen=fill_tcp_data_p(buf,plen,PSTR("<br>Preset: <select name=pr>"));
					if (_preset == 1 )
					plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0>Norm</option><option selected value=1>7I7O</option>"));
					else
					plen=fill_tcp_data_p(buf,plen,PSTR("<option value=0 selected>Norm</option><option value=1>7I7O</option>"));
					plen=fill_tcp_data_p(buf,plen,PSTR("</select><br>"));
					plen=fill_tcp_data_p(buf,plen,PSTR("T check: <input type=checkbox name=tc value=1"));
					//uint8_t temp_checked = eeprom_read_byte(&ee_temp_check);
					//if ( temp_checked == 1 )
					if ( eeprom_read_byte(&ee_temp_check) == 1 )
					plen=fill_tcp_data_p(buf,plen,PSTR(" checked"));
					plen=fill_tcp_data_p(buf,plen,PSTR("><br>"));

					plen=fill_tcp_data_p(buf,plen,PSTR("Alarm T: <input type=text name=at size=3 maxlength=3 value="));
					if ( _alarm_temp != 255 && _alarm_temp > 0 )
					{
						itoa (_alarm_temp, temp, 10);
						plen=fill_tcp_data(buf,plen,temp);
					}
					plen=fill_tcp_data_p(buf,plen,PSTR(">"));

					sprintf(temp,"%d", readchiptemp());
					plen=fill_tcp_data_p(buf,plen,PSTR("<br>Cur T: "));
					plen=fill_tcp_data(buf,plen,temp);
				}
				else if ( cf_mode == 2 )
				{
					plen=fill_tcp_data_p(buf,plen,PSTR("2>"));

					plen=fill_tcp_data_p(buf,plen,PSTR("Megad-ID: <input type=text name=mdid maxlength=5 size=5 value=\""));
					if ( _mdid[0] != (char)255 && strlen(_mdid) > 0 )
					plen=fill_tcp_data(buf,plen,_mdid);

					plen=fill_tcp_data_p(buf,plen,PSTR("\"><br>"));

					plen=fill_tcp_data_p(buf,plen,PSTR("srv loop: <input type=checkbox name=sl value=1"));
					if ( eeprom_read_byte(&ee_srv_loop) == 1 )
					plen=fill_tcp_data_p(buf,plen,PSTR(" checked"));
					plen=fill_tcp_data_p(buf,plen,PSTR(">"));

				}

				plen=fill_tcp_data_p(buf,plen,PSTR("<br><input type=submit value=Save></form>"));

				// ======================================================
				//client_browse_url(PSTR("/test-http.php"),"","",&browserresult_callback);


			}

		}

		else if (find_key_val((char *)&(buf[dat_p+4]),gStrbuf,25,"fwup"))
		{
			eeprom_write_byte ((void*)ee_fw_up, 1);
			reset_flag = 2;
			//Bootloader();
		}

		else
		{
		        plen=http200ok();

			if ( strlen(srv_cmd) > 0 )
			plen=fill_tcp_data_p(buf,plen,PSTR("Done"));
			else
			{

				eeprom_read_block (_port_type, &ee_port_type, IO_SIZE);

				plen=fill_tcp_data_p(buf,plen,PSTR("MegaD-328 by <a href=http://ab-log.ru>ab-log.ru</a> (fw: "));
				plen=fill_tcp_data_p(buf,plen,PSTR(MEGADFWV));
				plen=fill_tcp_data_p(buf,plen,PSTR(")<br><a href=/"));
				plen=fill_tcp_data(buf,plen,password);
				plen=fill_tcp_data_p(buf,plen,PSTR("/?cf=1>Config</a><br>-- Ports --"));

				//char pnum[3];
				for ( i = 0; i < IO_SIZE; i++ )
				{
					port_letter = aio[i][0];
					port_num = atoi(&aio[i][1]);

					plen=fill_tcp_data_p(buf,plen,PSTR("<br><a href=/"));
					plen=fill_tcp_data(buf,plen,password);
					plen=fill_tcp_data_p(buf,plen,PSTR("/?pt="));

					//char temp[2];

					snprintf_P(temp,sizeof(temp),PSTR("%d"),i);
					plen=fill_tcp_data(buf,plen, temp);

					if ( port_letter == 'A' )
					{
						plen=fill_tcp_data_p(buf,plen,PSTR(">"));
						plen=fill_tcp_data(buf,plen,aio[i]);
					}
					else
					{
						plen=fill_tcp_data_p(buf,plen,PSTR(">P"));
						plen=fill_tcp_data(buf,plen,temp);
					}

					plen=fill_tcp_data_p(buf,plen,PSTR(" - "));

					if ( _port_type[i] == 255 && port_letter != 'A' )
					plen=fill_tcp_data_p(buf,plen,PSTR("NC"));
					else if ( _port_type[i] == 1 )
					plen=fill_tcp_data_p(buf,plen,PSTR("OUT"));
					else if ( _port_type[i] == 0 )
					plen=fill_tcp_data_p(buf,plen,PSTR("IN"));
					else if ( _port_type[i] == 2 || port_letter == 'A' )
					plen=fill_tcp_data_p(buf,plen,PSTR("ADC"));
					else if ( _port_type[i] == 3 )
					plen=fill_tcp_data_p(buf,plen,PSTR("DSen"));

					plen=fill_tcp_data_p(buf,plen,PSTR("</a>"));
				}
			}

		}
		
		SENDTCP:

                LEDON;
                www_server_reply(buf,plen); // send data
		LEDOFF;

		if ( reset_flag == 1 )
		RESET();

		if ( reset_flag == 2 )
		Bootloader();

		continue;

		UDP:
		continue;
	}

        return (0);
}
